//
// Created by martin on 04.02.22.
//

#ifndef ANIMTESTBED_ANIMGRAPHRESOURCE_H
#define ANIMTESTBED_ANIMGRAPHRESOURCE_H

#include <cstring>
#include <iostream>
#include <map>
#include <string>
#include <type_traits>
#include <vector>

#include "SyncTrack.h"

enum class SocketType {
  SocketTypeUndefined,
  SocketTypeBool,
  SocketTypeAnimation,
  SocketTypeFloat,
  SocketTypeVec3,
  SocketTypeQuat,
  SocketTypeString
};

struct AnimData {
  float m_bone_transforms[16];
};

typedef float Vec3[3];
typedef float Quat[4];

inline int GenerateInputAttributeId(int node_id, int input_index) {
  return ((input_index + 1) << 14) + node_id;
}

inline void
SplitInputAttributeId(int attribute_id, int* node_id, int* input_index) {
  *node_id = attribute_id & ((1 << 14) - 1);
  *input_index = (attribute_id >> 14) - 1;
}

inline int GenerateOutputAttributeId(int node_id, int output_index) {
  return ((output_index + 1) << 23) + node_id;
}

inline void
SplitOutputAttributeId(int attribute_id, int* node_id, int* output_index) {
  *node_id = attribute_id & ((1 << 14) - 1);
  *output_index = (attribute_id >> 23) - 1;
}

struct AnimNodeResource;

struct Socket {
  std::string m_name;
  SocketType m_type = SocketType::SocketTypeUndefined;
  union SocketValue {
    void* ptr;
    void** ptr_ptr;
  };
  SocketValue m_value = {nullptr};
  size_t m_type_size = 0;
};

struct AnimNode;
struct NodeSocketAccessorBase;

struct AnimNodeResource {
  std::string m_name;
  std::string m_type_name;
  AnimNode* m_anim_node = nullptr;
  NodeSocketAccessorBase* m_socket_accessor = nullptr;

  float m_position[2] = {0.f, 0.f};
};

struct AnimNode {
  virtual ~AnimNode(){};

  std::string m_name;
  std::string m_node_type_name;
  bool m_is_time_synced;
  float m_time_now;
  float m_time_last;
  SyncTrack m_sync_track;
};

struct NodeSocketAccessorBase {
  NodeSocketAccessorBase() {}
  virtual ~NodeSocketAccessorBase() {}

  Socket* FindSocket(std::vector<Socket>& sockets, const std::string& name) {
    Socket* result = nullptr;
    for (size_t i = 0, n = sockets.size(); i < n; i++) {
      if (sockets[i].m_name == name) {
        result = &sockets[i];
        break;
      }
    }

    return result;
  }

  const Socket* FindSocket(
      const std::vector<Socket>& sockets,
      const std::string& name) const {
    const Socket* result = nullptr;
    for (size_t i = 0, n = sockets.size(); i < n; i++) {
      if (sockets[i].m_name == name) {
        result = &sockets[i];
        break;
      }
    }

    return result;
  }

  SocketType GetSocketType(
      const std::vector<Socket>& sockets,
      const std::string& name) {
    const Socket* socket = FindSocket(sockets, name);
    if (socket == nullptr) {
      return SocketType::SocketTypeUndefined;
    }
    return socket->m_type;
  }

  size_t GetSocketIndex(
      const std::vector<Socket>& sockets,
      const std::string& name) const {
    for (size_t i = 0, n = sockets.size(); i < n; i++) {
      std::cout << i << ", " << sockets[i].m_name << std::endl;
      if (sockets[i].m_name == name) {
        return i;
      }
    }

    return -1;
  }

  template <typename T>
  bool RegisterSocket(
      std::vector<Socket>& sockets,
      const std::string& name,
      T* value_ptr) {
    Socket* socket = FindSocket(sockets, name);
    if (socket != nullptr) {
      std::cerr << "Socket " << name << " already registered." << std::endl;
      return false;
    }

    sockets.push_back(Socket());
    socket = &sockets[sockets.size() - 1];
    socket->m_name = name;
    socket->m_type_size = sizeof(T);

    if constexpr (std::is_same<T, float>::value) {
      socket->m_type = SocketType::SocketTypeFloat;
    } else if constexpr (std::is_same<T, bool>::value) {
      socket->m_type = SocketType::SocketTypeBool;
    } else if constexpr (std::is_same<T, Vec3>::value) {
      socket->m_type = SocketType::SocketTypeVec3;
    } else if constexpr (std::is_same<T, Quat>::value) {
      socket->m_type = SocketType::SocketTypeQuat;
    } else if constexpr (std::is_same<T, AnimData>::value) {
      socket->m_type = SocketType::SocketTypeAnimation;
    } else if constexpr (std::is_same<T, std::string>::value) {
      socket->m_type = SocketType::SocketTypeString;
    } else if constexpr (std::is_same<T, float*>::value) {
      socket->m_type = SocketType::SocketTypeFloat;
    } else if constexpr (std::is_same<T, bool*>::value) {
      socket->m_type = SocketType::SocketTypeBool;
    } else if constexpr (std::is_same<T, Vec3*>::value) {
      socket->m_type = SocketType::SocketTypeVec3;
    } else if constexpr (std::is_same<T, Quat*>::value) {
      socket->m_type = SocketType::SocketTypeQuat;
    } else if constexpr (std::is_same<T, AnimData*>::value) {
      socket->m_type = SocketType::SocketTypeAnimation;
    } else if constexpr (std::is_same<T, std::string*>::value) {
      socket->m_type = SocketType::SocketTypeString;
    } else {
      std::cerr << "Cannot register socket, invalid type." << std::endl;
      return false;
    }

    socket->m_value.ptr = value_ptr;
    return true;
  }

  template <typename T>
  bool RegisterProperty(const std::string& name, T* value) {
    return RegisterSocket(m_properties, name, value);
  }
  template <typename T>
  bool SetProperty(const std::string& name, T value) {
    return SetSocketValue(m_properties, name, value);
  }
  template <typename T>
  T GetProperty(const std::string& name, T value) {
    return GetSocketValue(m_properties, name, value);
  }
  SocketType GetPropertyType(const std::string& name) {
    return GetSocketType(m_properties, name);
  }

  template <typename T>
  bool RegisterInput(const std::string& name, T* value) {
    return RegisterSocket(m_inputs, name, value);
  }
  template <typename T>
  T* GetInput(const std::string& name, T* value) {
    return GetSocketValue(m_inputs, name, value);
  }
  Socket* FindInputSocket(const std::string& name) {
    return FindSocket(m_inputs, name);
  }
  SocketType GetInputType(const std::string& name) {
    return GetSocketType(m_inputs, name);
  }
  size_t GetInputIndex(const std::string& name) {
    return GetSocketIndex(m_inputs, name);
  }

  template <typename T>
  bool RegisterOutput(const std::string& name, T** value) {
    return RegisterSocket(m_outputs, name, value);
  }
  SocketType GetOutputType(const std::string& name) {
    return GetSocketType(m_outputs, name);
  }
  Socket* FindOutputSocket(const std::string& name) {
    return FindSocket(m_outputs, name);
  }
  size_t GetOutputIndex(const std::string& name) {
    return GetSocketIndex(m_outputs, name);
  }

  std::vector<Socket> m_properties;
  std::vector<Socket> m_inputs;
  std::vector<Socket> m_outputs;
};

template <typename T>
struct NodeSocketAccessor : public NodeSocketAccessorBase {
  virtual ~NodeSocketAccessor() {}
};

struct NodeRegistry {
  AnimNode* createNode(const std::string& node_type);
  NodeSocketAccessorBase* createNodeAccessor(
      const std::string& node_type,
      AnimNode* node);
};

struct BlendTreeNode : public AnimNode {};

template <>
struct NodeSocketAccessor<BlendTreeNode> : public NodeSocketAccessorBase {
  NodeSocketAccessor(AnimNode* node_) {}
};

struct Blend2Node : public AnimNode {
  AnimData m_input0;
  AnimData m_input1;
  AnimData* m_output = nullptr;
  float m_blend_weight = 0.f;
  bool m_sync_blend = false;
};

template <>
struct NodeSocketAccessor<Blend2Node> : public NodeSocketAccessorBase {
  NodeSocketAccessor(AnimNode* node_) {
    Blend2Node* node = dynamic_cast<Blend2Node*>(node_);
    RegisterInput("Input0", &node->m_input0);
    RegisterInput("Input1", &node->m_input1);
    RegisterInput("Weight", &node->m_blend_weight);

    RegisterOutput("Output", &node->m_output);

    RegisterProperty("Sync", &node->m_sync_blend);
  }
};

struct SpeedScaleNode : public AnimNode {
  AnimData m_input;
  AnimData* m_output = nullptr;
  float m_speed_scale = 0.f;
};

template <>
struct NodeSocketAccessor<SpeedScaleNode> : public NodeSocketAccessorBase {
  NodeSocketAccessor(AnimNode* node_) {
    SpeedScaleNode* node = dynamic_cast<SpeedScaleNode*>(node_);
    RegisterInput("SpeedScale", &node->m_speed_scale);
    RegisterInput("Input", &node->m_input);

    RegisterOutput("Output", &node->m_output);
  }
};

struct AnimSamplerNode : public AnimNode {
  AnimData* m_output = nullptr;
  std::string m_filename;
};

template <>
struct NodeSocketAccessor<AnimSamplerNode> : public NodeSocketAccessorBase {
  NodeSocketAccessor(AnimNode* node_) {
    AnimSamplerNode* node = dynamic_cast<AnimSamplerNode*>(node_);
    RegisterOutput("Output", &node->m_output);

    RegisterProperty("Filename", &node->m_filename);
  }
};

struct AnimGraphConnection {
  int m_source_node_index;
  int m_source_socket_index;
  int m_target_node_index;
  int m_target_socket_index;
};

struct AnimGraphResource {
  std::string m_name;
  std::vector<AnimNodeResource> m_nodes;
  std::vector<AnimGraphConnection> m_connections;

  ~AnimGraphResource() {
    for (size_t i = 0, n = m_nodes.size(); i < n; i++) {
      delete m_nodes[i].m_anim_node;
      delete m_nodes[i].m_socket_accessor;
    }
  }

  AnimGraphResource() { clear(); }

  void clear();
  bool saveToFile(const char* filename) const;
  bool loadFromFile(const char* filename);

  size_t addNode(AnimNodeResource node_resource) {
    m_nodes.push_back(node_resource);
    return m_nodes.size() - 1;
  }
};

static inline AnimNode* AnimNodeFactory(const std::string& name) {
  AnimNode* result;
  if (name == "Blend2") {
    result = new Blend2Node;
  } else if (name == "SpeedScale") {
    result = new SpeedScaleNode;
  } else if (name == "AnimSampler") {
    result = new AnimSamplerNode;
  } else if (name == "BlendTree") {
    result = new BlendTreeNode;
  }

  if (result != nullptr) {
    result->m_node_type_name = name;
    return result;
  }

  std::cerr << "Invalid node type: " << name << std::endl;
  return nullptr;
}

static inline NodeSocketAccessorBase* AnimNodeAccessorFactory(
    const std::string& node_type_name,
    AnimNode* node) {
  if (node_type_name == "Blend2") {
    return new NodeSocketAccessor<Blend2Node>(node);
  } else if (node_type_name == "SpeedScale") {
    return new NodeSocketAccessor<SpeedScaleNode>(node);
  } else if (node_type_name == "AnimSampler") {
    return new NodeSocketAccessor<AnimSamplerNode>(node);
  } else if (node_type_name == "BlendTree") {
    return new NodeSocketAccessor<BlendTreeNode>(node);
  } else {
    std::cerr << "Invalid node type name " << node_type_name << "."
              << std::endl;
  }
  return nullptr;
}

static inline AnimNodeResource AnimNodeResourceFactory(
    const std::string& node_type_name) {
  AnimNodeResource result;
  result.m_type_name = node_type_name;
  result.m_anim_node = AnimNodeFactory(node_type_name);
  result.m_socket_accessor =
      AnimNodeAccessorFactory(node_type_name, result.m_anim_node);
  return result;
}

struct AnimGraph {
  void UpdateTime(float dt);
  void Evaluate();

  AnimData m_local_transforms;
  std::vector<AnimNode*> m_nodes;
  std::vector<std::vector<AnimNode*> > m_node_inputs;
  NodeSocketAccessorBase* m_socket_accessor;
  char* m_input_buffer = nullptr;
  char* m_output_buffer = nullptr;

  AnimNode* getAnimNode(const char* name) {
    for (size_t i = 0; i < m_nodes.size(); i++) {
      if (m_nodes[i]->m_name == name) {
        return m_nodes[i];
      }
    }

    return nullptr;
  }

  size_t getAnimNodeIndex(AnimNode* node) {
    for (size_t i = 0, n = m_nodes.size(); i < n; i++) {
      if (m_nodes[i] == node) {
        return i;
      }
    }

    std::cerr << "Could not find node " << node << "." << std::endl;
    return -1;
  }

  static AnimGraph createFromResource(const AnimGraphResource& resource) {
    AnimGraph result;

    // create nodes
    for (int i = 0; i < resource.m_nodes.size(); i++) {
      const AnimNodeResource& node_resource = resource.m_nodes[i];
      AnimNode* node = AnimNodeFactory(node_resource.m_type_name.c_str());
      node->m_name = node_resource.m_name;
      node->m_node_type_name = node_resource.m_type_name;
      result.m_nodes.push_back(node);
      result.m_node_inputs.push_back(std::vector<AnimNode*>());
    }

    // Prepare graph inputs
    const AnimNodeResource& graph_node = resource.m_nodes[0];
    result.m_socket_accessor = AnimNodeAccessorFactory("BlendTree", result.m_nodes[0]);

    int input_block_size = 0;
    const std::vector<Socket>& graph_inputs =
        graph_node.m_socket_accessor->m_inputs;
    for (int i = 0; i < graph_inputs.size(); i++) {
      input_block_size += graph_inputs[i].m_type_size;
    }
    result.m_input_buffer = new char[input_block_size];
    memset(result.m_input_buffer, 0, input_block_size);
    int input_block_offset = 0;
    for (int i = 0; i < graph_inputs.size(); i++) {
      if (graph_inputs[i].m_type == SocketType::SocketTypeAnimation) {
        result.m_socket_accessor->RegisterInput<AnimData>(graph_inputs[i].m_name, static_cast<AnimData*>( (void*) &result.m_input_buffer[input_block_offset]));
      }
      input_block_offset += graph_inputs[i].m_type_size;
    }

    // connect the nodes
    for (int i = 0; i < resource.m_connections.size(); i++) {
      const AnimGraphConnection& connection = resource.m_connections[i];
      std::string source_node_type = "";
      std::string target_node_type = "";
      std::string source_node_name = "";
      std::string target_node_name = "";
      AnimNode* source_node = nullptr;
      AnimNode* target_node = nullptr;
      NodeSocketAccessorBase* source_node_accessor = nullptr;
      NodeSocketAccessorBase* target_node_accessor = nullptr;
      SocketType source_type;
      SocketType target_type;

      if (connection.m_source_node_index >= 0) {
        source_node = result.m_nodes[connection.m_source_node_index];
        source_node_name = source_node->m_name;
        source_node_type = source_node->m_node_type_name;
        source_node_accessor =
            AnimNodeAccessorFactory(source_node_type, source_node);
      }

      if (connection.m_target_node_index >= 0) {
        target_node = result.m_nodes[connection.m_target_node_index];
        target_node_name = target_node->m_name;
        target_node_type = target_node->m_node_type_name;
        if (connection.m_target_node_index == 0) {
          target_node_accessor = result.m_socket_accessor;
        } else {
          target_node_accessor =
              AnimNodeAccessorFactory(target_node_type, target_node);
        }
      }

      assert(source_node != nullptr);
      assert(target_node != nullptr);

      if (connection.m_source_socket_index
          > source_node_accessor->m_outputs.size()) {
        std::cerr << "Invalid source socket index "
                  << connection.m_source_socket_index << ". Only "
                  << source_node_accessor->m_outputs.size()
                  << " output sockets found." << std::endl;
        continue;
      }

      if (connection.m_target_socket_index
          > target_node_accessor->m_inputs.size()) {
        std::cerr << "Invalid source socket index "
                  << connection.m_target_socket_index << ". Only "
                  << source_node_accessor->m_inputs.size()
                  << " input sockets found." << std::endl;
        continue;
      }

      if (source_node_accessor->m_outputs[connection.m_source_socket_index]
              .m_type
          != target_node_accessor->m_inputs[connection.m_target_socket_index]
                 .m_type) {
        std::cerr << "Invalid connection connecting nodes '" << source_node_name
                  << "' and '" << target_node_name << "'." << std::endl;
        return result;
      }

      Socket* source_socket =
          &source_node_accessor->m_outputs[connection.m_source_socket_index];
      Socket* target_socket =
          &target_node_accessor->m_inputs[connection.m_target_socket_index];
      if (source_socket->m_type != target_socket->m_type) {
        std::cerr << "Cannot connect sockets: invalid types!" << std::endl;
      }

      (*source_socket->m_value.ptr_ptr) = target_socket->m_value.ptr;

      size_t target_node_index = result.getAnimNodeIndex(target_node);
      result.m_node_inputs[target_node_index].push_back(source_node);
    }

    return result;
  }
};

#endif  //ANIMTESTBED_ANIMGRAPHRESOURCE_H
